#!/bin/bash -u

# CLOS like modifiers for bash functions.


_FUNCMOD_proxy () {

    local funcname=$1
    local list
    local var
    shift

    local args=( "$@" )

    var="_FUNCMOD_${funcname}_before"

    if [ "${!var:+set}" = set ]
    then

	eval 'list=( ${'"$var"'[@]} )'
	for func in "${list[@]}"; do
	    eval $func "$@"
	done
    fi

    var="_FUNCMOD_${funcname}_around"

    if [ "${!var:+set}" = set ]
    then

	eval "$var $@"

    fi

    var="_FUNCMOD_${funcname}_after"

    if [ "${!var:+set}" = set ]
    then

	eval 'list=( ${'"$var"'[@]} )'
	for func in "${list[@]}" ; do
	    eval $func "$@"
	done
    fi

}

_FUNCMOD_copy_func () {

    local src=$1
    local dst=$2

    local definition=$(declare -f $src)
    definition=${definition/#$src/$dst}

    eval "$definition"
}

_FUNCMOD_func_name() {

    local funcname=$1
    local cntr=_FUNCMOD_${funcname}_cnt

    local n=${!cntr}

    echo _FUNCMOD_${funcname}_${n}
}


_FUNCMOD_install_proxy() {

    local funcname=$1
    local cntr=_FUNCMOD_${funcname}_cnt

    [ "${!cntr:-}" != '' ] && return

    eval "$cntr=0"
    _FUNCMOD_copy_func $funcname _FUNCMOD_${funcname}_around
    eval $funcname'() { _FUNCMOD_proxy '"$funcname"' "$@" ; }'
}

before() {

    local funcname=$1
    _FUNCMOD_install_proxy $funcname

    local modifier=$2

    local list=_FUNCMOD_${funcname}_before

    eval "$list+=( $modifier )"
}

after() {

    local funcname=$1
    _FUNCMOD_install_proxy $funcname
    local modifier=$2

    local list=_FUNCMOD_${funcname}_after
    eval "$list+=( $modifier )"
}



around() {

    local funcname=$1

    _FUNCMOD_install_proxy $funcname

    local modifier=$2
    local cntr=_FUNCMOD_${funcname}_cnt

    eval "(( $cntr++ ))"

    local orig=$(_FUNCMOD_func_name $funcname)
    _FUNCMOD_copy_func  _FUNCMOD_${funcname}_around "$orig" 

    eval "_FUNCMOD_${funcname}_around"'() { '"$modifier"' '"$orig"' "$@" ; }'

}
